@@ -243,7 +243,7 @@ type connection_status = Ok | Bad
 type conninfo_option =
   {
     cio_keyword : string;
-    cio_envvar : string;
+    cio_envvar : string option;
     cio_compiled : string option;
     cio_val : string option;
     cio_label : string;
@@ -334,6 +334,12 @@ module Stub = struct
     connection -> string -> string array -> bool array -> result
     = "PQexecParams_stub"
 
+  external prepare : connection -> string -> string -> result = "PQprepare_stub"
+
+  external exec_prepared :
+    connection -> string -> string array -> bool array -> result
+    = "PQexecPrepared_stub"
+
   external describe_prepared :
     connection -> string -> result = "PQdescribePrepared_stub"
 
@@ -394,7 +400,7 @@ module Stub = struct
   external consume_input : connection -> int = "PQconsumeInput_stub" "noalloc"
   external is_busy : connection -> bool = "PQisBusy_stub" "noalloc"
   external flush : connection -> int = "PQflush_stub" "noalloc"
-  external socket : connection -> Unix.file_descr = "PQsocket_stub" "noalloc"
+  external socket : connection -> int = "PQsocket_stub" "noalloc"
   external request_cancel : connection -> string option = "PQCancel_stub"
 
 
@@ -446,9 +452,19 @@ module Stub = struct
   external lo_read :
     connection -> large_object -> string -> int -> int -> int = "lo_read_stub"
 
+  external lo_read_ba :
+    connection -> large_object ->
+    (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t ->
+    int -> int -> int = "lo_read_ba_stub"
+
   external lo_write :
     connection -> large_object -> string -> int -> int -> int = "lo_write_stub"
 
+  external lo_write_ba :
+    connection -> large_object ->
+    (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t ->
+    int -> int -> int = "lo_write_ba_stub"
+
   external lo_seek :
     connection -> large_object -> int -> seek_cmd -> int = "lo_lseek_stub"
 end
@@ -456,7 +472,56 @@ end
 
 (* Escaping *)
 
-external unescape_bytea : string -> string = "PQunescapeBytea_stub"
+external unescape_bytea_stub : string -> string = "PQunescapeBytea_stub"
+
+let unescape_bytea str =
+  let str_len = String.length str in
+  if str_len < 2 || str.[0] <> '\\' || str.[1] <> 'x' then
+    unescape_bytea_stub str
+  else
+    (* This is the new Postgresql 9.0 hex format for encoding bytea.  Older
+       clients do not recognize this format, but servers send it by default. *)
+    let rec count_hex ~pos ~count =
+      if pos = str_len then count
+      else
+        match str.[pos] with
+        | ' ' | '\t' | '\n' | '\r' | '\b' -> count_hex ~pos:(pos + 1) ~count
+        | '0'..'9' | 'a'..'f' | 'A'..'F' ->
+            let pos = pos + 1 in
+            if pos = str_len then
+              failwith "Postgresql.unescape_bytea: incomplete hex code"
+            else
+              begin match str.[pos] with
+              | '0'..'9' | 'a'..'f' | 'A'..'F' ->
+                  count_hex ~pos:(pos + 1) ~count:(count + 1)
+              | _ ->
+                  failwith "Postgresql.unescape_bytea: short hex code"
+              end
+        | _ -> failwith "Postgresql.unescape_bytea: illegal hex code"
+    in
+    let n_grps = count_hex ~pos:2 ~count:0 in
+    let res = String.create n_grps in
+    let rec res_loop ~str_pos ~res_pos =
+      if res_pos = n_grps then res
+      else
+        match str.[str_pos] with
+        | ' ' | '\t' | '\n' | '\r' | '\b' ->
+            res_loop ~str_pos:(str_pos + 1) ~res_pos
+        | c ->
+            let unhexdigit = function
+              | '0'..'9' as c -> Char.code c - Char.code '0'
+              | 'a'..'f' as c -> Char.code c - Char.code 'a' + 10
+              | 'A'..'F' as c -> Char.code c - Char.code 'A' + 10
+              | _ -> assert false
+            in
+            let high = unhexdigit c lsl 4 in
+            let str_pos = str_pos + 1 in
+            let low = unhexdigit str.[str_pos] in
+            let res_char = Char.chr (high + low) in
+            res.[res_pos] <- res_char;
+            res_loop ~str_pos:(str_pos + 1) ~res_pos:(res_pos + 1)
+    in
+    res_loop ~str_pos:2 ~res_pos:0
 
 
 (* Query results *)
@@ -730,6 +795,27 @@ object (self)
       raise (Error (Unexpected_status (stat, res#error, expect)))
     else res
 
+  method prepare stm_name query =
+    new result (
+      wrap_conn (fun conn ->
+        let r = Stub.prepare conn stm_name query in
+        if Stub.result_isnull r then signal_error conn
+        else r))
+
+  method exec_prepared
+    ?(expect = []) ?(params = [||]) ?(binary_params = [||]) stm_name =
+    let r =
+      wrap_conn (fun conn ->
+        let r = Stub.exec_prepared conn stm_name params binary_params in
+        if Stub.result_isnull r then signal_error conn
+        else r)
+    in
+    let res = new result r in
+    let stat = res#status in
+    if not (expect = []) && not (List.mem stat expect) then
+      raise (Error (Unexpected_status (stat, res#error, expect)))
+    else res
+
   method describe_prepared query =
     new result (
       wrap_conn (fun conn ->
@@ -804,12 +890,14 @@ object (self)
         match r with
          | 0 -> f (Buffer.contents buf); Buffer.clear buf; line ()
          | 1 -> loop (Stub.getline conn s 0 len)
-         | _ -> f (Buffer.contents buf); self#endcopy
+         | _ -> f (Buffer.contents buf)
       and line () =
         let r = Stub.getline conn s 0 len in
-        if s.[0] = '\\' && s.[1] = '.' && s.[2] = '\000' then self#endcopy
-        else loop r in
-      line ())
+        if r < 3 || s.[0] <> '\\' || s.[1] <> '.' || s.[2] <> '\000' then
+          loop r
+      in
+      line ());
+    self#endcopy
 
   method copy_out_channel oc =
     self#copy_out (fun s -> output_string oc (s ^ "\n"))
@@ -839,7 +927,7 @@ object (self)
   method socket =
     wrap_conn (fun conn ->
       let s = Stub.socket conn in
-      if Obj.magic s = -1 then signal_error conn else s)
+      if s = -1 then signal_error conn else s)
 
   method request_cancel = request_cancel ()
 
@@ -877,6 +965,15 @@ object (self)
       let w = Stub.lo_write conn lo buf pos len in
       if w < len then signal_error conn)
 
+  method lo_write_ba ?(pos = 0) ?len buf lo =
+    let buf_len = Bigarray.Array1.dim buf in
+    let len = match len with Some len -> len | None -> buf_len - pos in
+    if len < 0 || pos < 0 || pos + len > buf_len then
+      invalid_arg "Postgresql.connection#lo_write_ba";
+    wrap_conn (fun conn ->
+      let w = Stub.lo_write_ba conn lo buf pos len in
+      if w < len then signal_error conn)
+
   method lo_read lo ?(pos = 0) ?len buf =
     let buf_len = String.length buf in
     let len = match len with Some len -> len | None -> buf_len - pos in
@@ -887,6 +984,16 @@ object (self)
       if read = -1 then signal_error conn;
       read)
 
+  method lo_read_ba lo ?(pos = 0) ?len buf =
+    let buf_len = Bigarray.Array1.dim buf in
+    let len = match len with Some len -> len | None -> buf_len - pos in
+    if len < 0 || pos < 0 || pos + len > buf_len then
+      invalid_arg "Postgresql.connection#lo_read_ba";
+    wrap_conn (fun conn ->
+      let read = Stub.lo_read_ba conn lo buf pos len in
+      if read = -1 then signal_error conn;
+      read)
+
   method lo_seek ?(pos = 0) ?(whence = SEEK_SET) lo =
     wrap_conn (fun conn ->
       if Stub.lo_seek conn lo pos whence < 0 then signal_error conn)
