@@ -169,8 +169,11 @@ exception Oid of oid
 
 (** {6 Utility functions}*)
 
-external unescape_bytea : string -> string = "PQunescapeBytea_stub"
-(** [unescape_bytea str] unescapes binary string [str]. *)
+val unescape_bytea : string -> string
+(** [unescape_bytea str] unescapes binary string [str].  This function
+    supports the new hex format for encoding bytea strings (introduced
+    in Postgresql 9.0) even if the local libpq library is from an
+    older version. *)
 
 external ftype_of_oid : oid -> ftype = "ftype_of_oid_stub"
 (** [ftype_of_oid oid] converts [oid] to an [ftype]. *)
@@ -356,7 +359,7 @@ type connection_status = Ok | Bad
 type conninfo_option =
   {
     cio_keyword : string;  (** Keyword of option *)
-    cio_envvar : string;  (** Fallback environment variable name *)
+    cio_envvar : string option;  (** Fallback environment variable name *)
     cio_compiled : string option;  (** Fallback compiled in default value *)
     cio_val : string option;  (** Current value of option, or NULL *)
     cio_label : string;  (** Label for field in connect dialog *)
@@ -530,6 +533,17 @@ object
       @raise Error if there is an unexpected result status.
   *)
 
+  method prepare : string -> string -> result
+  (** [prepare stm_name query] creates a prepared query named [stm_name]
+      which will execute the query or command [query] when passed to
+      [#exec_prepared]. *)
+
+  method exec_prepared :
+    ?expect : result_status list -> ?params : string array ->
+    ?binary_params : bool array -> string -> result
+  (** [exec_prepared ?expect ?params ?binary_params stm_name] acts as
+      [#exec], except executes the prepared query [stm_name]. *)
+
   method describe_prepared : string -> result
   (** [#describe_prepared stm_name] submits a request to obtain
       information about the specified prepared statement, and waits for
@@ -699,9 +713,9 @@ object
       @raise Error if there is a connection error.
   *)
 
-  method socket : Unix.file_descr
+  method socket : int
   (** [socket] obtains the file descriptor for the backend connection
-      socket.
+      socket as an integer.
 
       @raise Error if there is a connection error.
   *)
@@ -751,6 +765,12 @@ object
       @raise Error if [len] bytes could not be written.
       @raise Error if there is a connection error. *)
 
+  method lo_write_ba :
+    ?pos : int -> ?len : int ->
+    (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t ->
+    large_object -> unit
+  (** As [lo_write], but performs a zero-copy write from the given Bigarray. *)
+
   method lo_read : large_object -> ?pos : int -> ?len : int -> string -> int
   (** [lo_read lo ?pos ?len buf] reads [len] bytes from large object [lo]
       to buffer [buf] starting at position [pos].
@@ -763,6 +783,11 @@ object
       @raise Error if there is a connection error.
   *)
 
+  method lo_read_ba :
+    large_object -> ?pos : int -> ?len : int ->
+    (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t -> int
+  (** As [lo_read], but performs a zero-copy read into the given Bigarray. *)
+
   method lo_seek : ?pos : int -> ?whence : seek_cmd -> large_object -> unit
   (** [lo_seek ?pos ?whence lo] seeks read/write position [pos] in
       large object [lo] relative to the start, current read/write
