@@ -45,6 +45,7 @@
 #include <caml/callback.h>
 #include <caml/signals.h>
 #include <caml/fail.h>
+#include <caml/bigarray.h>
 
 #include <libpq-fe.h>
 #include <libpq/libpq-fs.h>
@@ -209,11 +210,14 @@ static inline void np_decr_refcount(np_callback *c)
      PQgetssl: the SSL structure used in the connection
 */
 
-#define get_conn(v) ((PGconn *) Field(v, 1))
-#define set_conn(v, conn) (Field(v, 1) = (value) conn)
+#define get_conn(v) ((PGconn *) Field(v, 0))
+#define set_conn(v, conn) (Field(v, 0) = (value) conn)
 
-#define get_conn_cb(v) ((np_callback *) Field(v, 2))
-#define set_conn_cb(v, cb) (Field(v, 2) = (value) cb)
+#define get_conn_cb(v) ((np_callback *) Field(v, 1))
+#define set_conn_cb(v, cb) (Field(v, 1) = (value) cb)
+
+#define get_cancel_obj(v) ((PGcancel *) Field(v, 2))
+#define set_cancel_obj(v, cancel) (Field(v, 2) = (value) cancel)
 
 CAMLprim value PQconn_isnull(value v_conn)
 {
@@ -222,13 +226,15 @@ CAMLprim value PQconn_isnull(value v_conn)
 
 static inline void free_conn(value v_conn)
 {
-  PGconn *conn;
-  np_decr_refcount(get_conn_cb(v_conn));
-  set_conn_cb(v_conn, NULL);
-  conn = get_conn(v_conn);
+  PGconn *conn = get_conn(v_conn);
   if (conn) {
+    PGcancel *cancel = get_cancel_obj(v_conn);
+    set_cancel_obj(v_conn, NULL);
+    np_decr_refcount(get_conn_cb(v_conn));
+    set_conn_cb(v_conn, NULL);
     set_conn(v_conn, NULL);
     caml_enter_blocking_section();
+      PQfreeCancel(cancel);
       PQfinish(conn);
     caml_leave_blocking_section();
   }
@@ -239,21 +245,24 @@ CAMLprim value PQconnectdb_stub(value v_conn_info)
   PGconn *conn;
   value v_conn;
 
-  int len = caml_string_length(v_conn_info) + 1;
+  size_t len = caml_string_length(v_conn_info) + 1;
+  PGcancel *cancel;
   char *conn_info = caml_stat_alloc(len);
   memcpy(conn_info, String_val(v_conn_info), len);
 
   caml_enter_blocking_section();
     conn = PQconnectdb(conn_info);
+    cancel = PQgetCancel(conn);
     free(conn_info);
   caml_leave_blocking_section();
 
   /* One may raise this 30 to 500 for instance if the program takes
      responsibility of closing connections */
-  v_conn = caml_alloc_final(3, free_conn, 1, 30);
+  v_conn = caml_alloc_small(3, Abstract_tag);
 
   set_conn(v_conn, conn);
   set_conn_cb(v_conn, NULL);
+  set_cancel_obj(v_conn, cancel);
 
   return v_conn;
 }
@@ -277,34 +286,44 @@ CAMLprim value PQreset_stub(value v_conn)
 CAMLprim value PQconndefaults_stub(value __unused v_unit)
 {
   CAMLparam0();
-  CAMLlocal1(v_res);
+  CAMLlocal2(v_res, v_el);
   PQconninfoOption *cios = PQconndefaults(), *p = cios;
   int i, j, n;
 
   while (p->keyword != NULL) p++;
 
   n = p - cios;
+  p = cios;
   v_res = caml_alloc_tuple(n);
 
-  for (i = 0; i < n; i++, cios++) {
-    value v_el = caml_alloc_small(7, 0);
-    for (j = 0; j < 7; j++) { Field(v_el, j) = v_None; };
+  for (i = 0; i < n; i++, p++) {
+    value v_field;
+    v_el = caml_alloc_small(7, 0);
+    for (j = 0; j < 7; j++) Field(v_el, j) = v_None;
     Store_field(v_res, i, v_el);
-    Field(v_el, 0) = caml_copy_string(cios->keyword);
-    caml_modify(&Field(v_el, 1), caml_copy_string(cios->envvar));
-    if (cios->compiled) {
-      value v_Some = make_some(caml_copy_string(cios->compiled));
-      caml_modify(&Field(v_el, 2), v_Some);
+    v_field = caml_copy_string(p->keyword);
+    Field(v_el, 0) = v_field;
+    if (p->envvar) {
+      v_field = make_some(caml_copy_string(p->envvar));
+      caml_modify(&Field(v_el, 1), v_field);
+    }
+    if (p->compiled) {
+      v_field = make_some(caml_copy_string(p->compiled));
+      caml_modify(&Field(v_el, 2), v_field);
     };
-    if (cios->val) {
-      value v_Some = make_some(caml_copy_string(cios->val));
-      caml_modify(&Field(v_el, 3), v_Some);
+    if (p->val) {
+      v_field = make_some(caml_copy_string(p->val));
+      caml_modify(&Field(v_el, 3), v_field);
     };
-    caml_modify(&Field(v_el, 4), caml_copy_string(cios->label));
-    caml_modify(&Field(v_el, 5), caml_copy_string(cios->dispchar));
-    caml_modify(&Field(v_el, 6), Val_int(cios->dispsize));
+    v_field = caml_copy_string(p->label);
+    caml_modify(&Field(v_el, 4), v_field);
+    v_field = caml_copy_string(p->dispchar);
+    caml_modify(&Field(v_el, 5), v_field);
+    caml_modify(&Field(v_el, 6), Val_int(p->dispsize));
   };
 
+  PQconninfoFree(cios);
+
   CAMLreturn(v_res);
 }
 
@@ -363,13 +382,13 @@ noalloc_conn_info(PQbackendPID, Val_int)
   CAMLprim value fun##_stub(value v_res, value v_field_num) \
   { \
     CAMLparam1(v_res); \
-    CAMLreturn(ret(fun(get_res(v_res), Int_val(v_field_num)))); \
+    CAMLreturn(ret(fun(get_res(v_res), Long_val(v_field_num)))); \
   }
 
 #define noalloc_fieldnum_info(fun, ret) \
   CAMLprim value fun##_stub(value v_res, value v_field_num) \
   { \
-    return ret(fun(get_res(v_res), Int_val(v_field_num))); \
+    return ret(fun(get_res(v_res), Long_val(v_field_num))); \
   }
 
 #define field_info(fun, ret) \
@@ -377,13 +396,14 @@ noalloc_conn_info(PQbackendPID, Val_int)
   { \
     CAMLparam1(v_res); \
     CAMLreturn( \
-      ret(fun(get_res(v_res), Int_val(v_tup_num), Int_val(v_field_num)))); \
+      ret(fun(get_res(v_res), Long_val(v_tup_num), Long_val(v_field_num)))); \
   }
 
 #define noalloc_field_info(fun, ret) \
   CAMLprim value fun##_stub(value v_res, value v_tup_num, value v_field_num) \
   { \
-    return ret(fun(get_res(v_res), Int_val(v_tup_num), Int_val(v_field_num))); \
+    return \
+      ret(fun(get_res(v_res), Long_val(v_tup_num), Long_val(v_field_num))); \
   }
 
 static inline void free_result(value v_res)
@@ -413,7 +433,7 @@ static inline value alloc_result(PGresult *res, np_callback *cb)
 }
 
 static inline void copy_binary_params(
-  value v_params, value v_binary_params, int nparams,
+  value v_params, value v_binary_params, size_t nparams,
   int **res_formats, int **res_lengths)
 {
   int i, nbinary, *lengths, *formats;
@@ -446,17 +466,17 @@ static inline void free_binary_params(int *formats, int *lengths)
   if (lengths != NULL) free(lengths);
 }
 
-static inline const char * const * copy_params(value v_params, int nparams)
+static inline const char * const * copy_params(value v_params, size_t nparams)
 {
   char **params;
-  int i;
+  size_t i;
   if (nparams == 0) return NULL;
   params = caml_stat_alloc(nparams * sizeof(char *));
   for (i = 0; i < nparams; i++) {
     value v_param = Field(v_params, i);
     if (v_param == *v_null_param) params[i] = NULL;
     else {
-      int param_len = caml_string_length(v_param) + 1;
+      size_t param_len = caml_string_length(v_param) + 1;
       params[i] = caml_stat_alloc(param_len);
       memcpy(params[i], String_val(v_param), param_len);
     }
@@ -464,19 +484,19 @@ static inline const char * const * copy_params(value v_params, int nparams)
   return (const char * const *) params;
 }
 
-static inline void free_params(const char * const *params, int nparams)
+static inline void free_params(const char * const *params, size_t nparams)
 {
-  int i;
+  size_t i;
   if (nparams == 0) return;
   for (i = 0; i < nparams; i++) caml_stat_free((char *) params[i]);
   free((char **) params);
 }
 
 static inline const char * const * copy_params_shallow(
-  value v_params, int nparams)
+  value v_params, size_t nparams)
 {
   char **params;
-  int i;
+  size_t i;
   if (nparams == 0) return NULL;
   params = caml_stat_alloc(nparams * sizeof(char *));
   for (i = 0; i < nparams; i++) {
@@ -486,7 +506,8 @@ static inline const char * const * copy_params_shallow(
   return (const char * const *) params;
 }
 
-static inline void free_params_shallow(const char * const *params, int nparams)
+static inline void free_params_shallow(
+  const char * const *params, size_t nparams)
 {
   if (nparams == 0) return;
   free((char **) params);
@@ -499,9 +520,9 @@ CAMLprim value PQexecParams_stub(
   PGconn *conn = get_conn(v_conn);
   np_callback *np_cb = get_conn_cb(v_conn);
   PGresult *res;
-  int len = caml_string_length(v_query) + 1;
+  size_t len = caml_string_length(v_query) + 1;
   char *query = caml_stat_alloc(len);
-  int nparams = Wosize_val(v_params);
+  size_t nparams = Wosize_val(v_params);
   const char * const *params = copy_params(v_params, nparams);
   int *formats, *lengths;
   copy_binary_params(v_params, v_binary_params, nparams, &formats, &lengths);
@@ -518,6 +539,66 @@ CAMLprim value PQexecParams_stub(
   CAMLreturn(alloc_result(res, np_cb));
 }
 
+#ifdef PG_OCAML_8_2
+CAMLprim value PQprepare_stub(value v_conn, value v_stm_name, value v_query)
+{
+  CAMLparam1(v_conn);
+  PGconn *conn = get_conn(v_conn);
+  np_callback *np_cb = get_conn_cb(v_conn);
+  PGresult *res;
+  size_t stm_name_len = caml_string_length(v_stm_name) + 1;
+  size_t query_len = caml_string_length(v_query) + 1;
+  char *stm_name = caml_stat_alloc(stm_name_len);
+  char *query = caml_stat_alloc(query_len);
+  memcpy(stm_name, String_val(v_stm_name), stm_name_len);
+  memcpy(query, String_val(v_query), query_len);
+  caml_enter_blocking_section();
+    res = PQprepare(conn, stm_name, query, 0, NULL);
+    free(stm_name);
+    free(query);
+  caml_leave_blocking_section();
+  CAMLreturn(alloc_result(res, np_cb));
+#else
+CAMLprim value PQprepare_stub(
+  value __unused v_conn, value __unused v_stm_name, value __unused v_query)
+{
+  caml_failwith("Postgresql.prepare: not supported");
+  return Val_unit;
+#endif
+}
+
+#ifdef PG_OCAML_8_2
+CAMLprim value PQexecPrepared_stub(
+  value v_conn, value v_stm_name, value v_params, value v_binary_params)
+{
+  CAMLparam1(v_conn);
+  PGconn *conn = get_conn(v_conn);
+  np_callback *np_cb = get_conn_cb(v_conn);
+  PGresult *res;
+  size_t len = caml_string_length(v_stm_name) + 1;
+  char *stm_name = caml_stat_alloc(len);
+  size_t nparams = Wosize_val(v_params);
+  const char * const *params = copy_params(v_params, nparams);
+  int *formats, *lengths;
+  copy_binary_params(v_params, v_binary_params, nparams, &formats, &lengths);
+  memcpy(stm_name, String_val(v_stm_name), len);
+  caml_enter_blocking_section();
+    res = PQexecPrepared(conn, stm_name, nparams, params, lengths, formats, 0);
+    free(stm_name);
+    free_params(params, nparams);
+    free_binary_params(formats, lengths);
+  caml_leave_blocking_section();
+  CAMLreturn(alloc_result(res, np_cb));
+#else
+CAMLprim value PQexecPrepared_stub(
+  value __unused v_conn, value __unused v_stm_name, value __unused v_params,
+  value __unused v_binary_params)
+{
+  caml_failwith("Postgresql.exec_prepared: not supported");
+  return Val_unit;
+#endif
+}
+
 #ifdef PG_OCAML_8_2
 CAMLprim value PQdescribePrepared_stub(value v_conn, value v_query)
 {
@@ -525,7 +606,7 @@ CAMLprim value PQdescribePrepared_stub(value v_conn, value v_query)
   PGconn *conn = get_conn(v_conn);
   np_callback *np_cb = get_conn_cb(v_conn);
   PGresult *res;
-  int len = caml_string_length(v_query) + 1;
+  size_t len = caml_string_length(v_query) + 1;
   char *query = caml_stat_alloc(len);
   memcpy(query, String_val(v_query), len);
   caml_enter_blocking_section();
@@ -550,13 +631,13 @@ CAMLprim value PQresStatus_stub(value v_status)
 }
 
 res_info(PQresultErrorMessage, make_string)
-noalloc_res_info(PQntuples, Val_int)
-noalloc_res_info(PQnfields, Val_int)
+noalloc_res_info(PQntuples, Val_long)
+noalloc_res_info(PQnfields, Val_long)
 noalloc_res_info(PQbinaryTuples, Val_bool)
 fieldnum_info(PQfname, make_string)
 
 #ifdef PG_OCAML_8_2
-noalloc_res_info(PQnparams, Val_int)
+noalloc_res_info(PQnparams, Val_long)
 #else
 CAMLprim value PQnparams_stub(value __unused v_res)
 {
@@ -567,16 +648,16 @@ CAMLprim value PQnparams_stub(value __unused v_res)
 
 CAMLprim value PQfnumber_stub(value v_res, value v_field_name)
 {
-  return Val_int(PQfnumber(get_res(v_res), String_val(v_field_name)));
+  return Val_long(PQfnumber(get_res(v_res), String_val(v_field_name)));
 }
 
 noalloc_fieldnum_info(PQfformat, Val_int)
 noalloc_fieldnum_info(PQftype, Val_int)
-noalloc_fieldnum_info(PQfsize, Val_int)
+noalloc_fieldnum_info(PQfsize, Val_long)
 noalloc_fieldnum_info(PQfmod, Val_int)
 
 #ifdef PG_OCAML_8_2
-noalloc_fieldnum_info(PQparamtype, Val_int)
+noalloc_fieldnum_info(PQparamtype, Val_long)
 #else
 CAMLprim value
 PQparamtype_stub(value __unused v_res, value __unused v_field_num)
@@ -592,20 +673,20 @@ CAMLprim value PQgetvalue_stub(value v_res, value v_tup_num, value v_field_num)
   CAMLparam1(v_res);
   value v_str;
   PGresult *res = get_res(v_res);
-  int field_num = Int_val(v_field_num);
-  int tup_num = Int_val(v_tup_num);
+  size_t field_num = Long_val(v_field_num);
+  size_t tup_num = Long_val(v_tup_num);
   char *str = PQgetvalue(res, tup_num, field_num);
   if (PQfformat(res, field_num) == 0) v_str = make_string(str);
   else {
     /* Assume binary format! */
-    int len = PQgetlength(res, tup_num, field_num);
+    size_t len = PQgetlength(res, tup_num, field_num);
     v_str = len ? v_empty_string : caml_alloc_string(len);
     memcpy(String_val(v_str), str, len);
   }
   CAMLreturn(v_str);
 }
 
-noalloc_field_info(PQgetlength, Val_int)
+noalloc_field_info(PQgetlength, Val_long)
 noalloc_field_info(PQgetisnull, Val_bool)
 
 res_info(PQcmdStatus, make_string)
@@ -636,7 +717,7 @@ CAMLprim value PQsendQueryParams_stub(
 {
   PGconn *conn = get_conn(v_conn);
   const char *query = String_val(v_query);
-  int nparams = Wosize_val(v_params);
+  size_t nparams = Wosize_val(v_params);
   const char * const *params = copy_params_shallow(v_params, nparams);
   int *lengths, *formats, res;
   copy_binary_params(v_params, v_binary_params, nparams, &formats, &lengths);
@@ -665,14 +746,45 @@ noalloc_conn_info(PQconsumeInput, Val_int)
 noalloc_conn_info(PQisBusy, Val_bool)
 noalloc_conn_info(PQflush, Val_int)
 noalloc_conn_info(PQsocket, Val_int)
-noalloc_conn_info(PQrequestCancel, Val_int)
 
-CAMLprim value PQescapeString_stub(value v_from, value v_pos_from,
-                                   value v_to, value v_pos_to, value v_len)
+CAMLprim value PQCancel_stub(value v_conn)
 {
-  return Val_int(PQescapeString(String_val(v_to) + Int_val(v_pos_to),
-                                String_val(v_from) + Int_val(v_pos_from),
-                                Int_val(v_len)));
+  CAMLparam1(v_conn);
+  PGconn *conn = get_conn(v_conn);
+  if (conn == NULL) CAMLreturn(v_None);
+  else {
+    PGcancel *cancel = get_cancel_obj(v_conn);
+    char errbuf[256];
+    int res;
+    caml_enter_blocking_section();
+      res = PQcancel(cancel, errbuf, 256);
+    caml_leave_blocking_section();
+    if (res == 0) CAMLreturn(v_None);
+    else CAMLreturn(make_some(caml_copy_string(errbuf)));
+  }
+}
+
+CAMLprim value PQescapeStringConn_stub(
+  value v_conn, value v_from, value v_pos_from, value v_len)
+{
+  size_t len = Long_val(v_len);
+  size_t to_len = len + len + 1;
+  char *buf = malloc(to_len);
+  int error;
+  size_t n_written =
+    PQescapeStringConn(
+      get_conn(v_conn),
+      buf, String_val(v_from) + Long_val(v_pos_from),
+      len, &error);
+  if (error) {
+    free(buf);
+    caml_failwith("Postgresql.escape_string_conn: failed to escape string");
+  } else {
+    value v_res = caml_alloc_string(n_written);
+    memcpy(String_val(v_res), buf, n_written);
+    free(buf);
+    return v_res;
+  }
 }
 
 CAMLprim value PQescapeByteaConn_stub(
@@ -682,8 +794,8 @@ CAMLprim value PQescapeByteaConn_stub(
   char *buf =
     (char *) PQescapeByteaConn(
       get_conn(v_conn),
-      (unsigned char *) String_val(v_from) + Int_val(v_pos_from),
-      Int_val(v_len), &len);
+      (unsigned char *) String_val(v_from) + Long_val(v_pos_from),
+      Long_val(v_len), &len);
   value v_res = caml_alloc_string(--len);
   memcpy(String_val(v_res), buf, len);
   PQfreemem(buf);
@@ -710,10 +822,10 @@ CAMLprim value PQunescapeBytea_stub(value v_from)
 CAMLprim value PQnotifies_stub(value v_conn)
 {
   CAMLparam1(v_conn);
+  CAMLlocal1(v_str);
   PGnotify *noti = PQnotifies(get_conn(v_conn));
 
   if (noti) {
-    CAMLlocal1(v_str);
     value v_pair;
     v_str = make_string(noti->relname);
     v_pair = caml_alloc_small(2, 0);
@@ -733,12 +845,12 @@ CAMLprim value PQgetline_stub(
   CAMLparam2(v_conn, v_buf);
   PGconn *conn = get_conn(v_conn);
   value v_res;
-  int len = Int_val(v_len);
+  size_t len = Long_val(v_len);
   char *buf = caml_stat_alloc(len);
   caml_enter_blocking_section();
     v_res = Val_int(PQgetline(conn, buf, len));
   caml_leave_blocking_section();
-  memcpy(String_val(v_buf) + Int_val(v_pos), buf, len);
+  memcpy(String_val(v_buf) + Long_val(v_pos), buf, len);
   free(buf);
   CAMLreturn(v_res);
 }
@@ -747,8 +859,8 @@ CAMLprim value PQgetlineAsync_stub(
   value v_conn, value v_buf, value v_pos, value v_len)
 {
   return Val_int(PQgetlineAsync(get_conn(v_conn),
-                                String_val(v_buf) + Int_val(v_pos),
-                                Int_val(v_len)));
+                                String_val(v_buf) + Long_val(v_pos),
+                                Long_val(v_len)));
 }
 
 CAMLprim value PQputline_stub(value v_conn, value v_buf)
@@ -756,7 +868,7 @@ CAMLprim value PQputline_stub(value v_conn, value v_buf)
   CAMLparam1(v_conn);
   PGconn *conn = get_conn(v_conn);
   value v_res;
-  int len = caml_string_length(v_buf) + 1;
+  size_t len = caml_string_length(v_buf) + 1;
   char *buf = caml_stat_alloc(len);
   memcpy(buf, String_val(v_buf), len);
   caml_enter_blocking_section();
@@ -772,9 +884,9 @@ CAMLprim value PQputnbytes_stub(
   CAMLparam1(v_conn);
   PGconn *conn = get_conn(v_conn);
   value v_res;
-  int len = Int_val(v_len);
+  size_t len = Long_val(v_len);
   char *buf = caml_stat_alloc(len);
-  memcpy(buf, String_val(v_buf) + Int_val(v_pos), len);
+  memcpy(buf, String_val(v_buf) + Long_val(v_pos), len);
   caml_enter_blocking_section();
     v_res = Val_int(PQputnbytes(conn, buf, len));
     free(buf);
@@ -798,8 +910,14 @@ CAMLprim value PQendcopy_stub(value v_conn)
 
 static inline void notice_ml(void *cb, const char *msg)
 {
-  value v_msg = make_string(msg);
-  caml_callback(((np_callback *) cb)->v_cb, v_msg);
+  value v_msg;
+  /* CR mmottl for mmottl: this is not reliable and can lead to segfaults,
+     because the runtime lock may already be held (but not usually).
+     A runtime feature is needed to fully support this. */
+  caml_leave_blocking_section();
+    v_msg = make_string(msg);
+    caml_callback(((np_callback *) cb)->v_cb, v_msg);
+  caml_enter_blocking_section();
 }
 
 CAMLprim value PQsetNoticeProcessor_stub(value v_conn, value v_cb)
@@ -841,32 +959,60 @@ CAMLprim value lo_read_stub(value v_conn, value v_fd,
   CAMLparam2(v_conn, v_buf);
   PGconn *conn = get_conn(v_conn);
   value v_res;
-  int len = Int_val(v_len);
+  size_t len = Long_val(v_len);
   char *buf = caml_stat_alloc(len);
   caml_enter_blocking_section();
-    v_res = Val_int(lo_read(conn, Int_val(v_fd), buf, len));
+    v_res = Val_long(lo_read(conn, Int_val(v_fd), buf, len));
   caml_leave_blocking_section();
-  memcpy(String_val(v_buf) + Int_val(v_pos), buf, len);
+  memcpy(String_val(v_buf) + Long_val(v_pos), buf, len);
   free(buf);
   CAMLreturn(v_res);
 }
 
+CAMLprim value lo_read_ba_stub(value v_conn, value v_fd,
+                               value v_buf, value v_pos, value v_len)
+{
+  CAMLparam2(v_conn, v_buf);
+  PGconn *conn = get_conn(v_conn);
+  value v_res;
+  size_t len = Long_val(v_len);
+  char *buf = ((char *) Caml_ba_data_val(v_buf)) + Long_val(v_pos);
+  caml_enter_blocking_section();
+    v_res = Val_long(lo_read(conn, Int_val(v_fd), buf, len));
+  caml_leave_blocking_section();
+  CAMLreturn(v_res);
+}
+
 CAMLprim value lo_write_stub(value v_conn, value v_fd,
                              value v_buf, value v_pos, value v_len)
 {
   CAMLparam1(v_conn);
   PGconn *conn = get_conn(v_conn);
   value v_res;
-  int len = Int_val(v_len);
+  size_t len = Long_val(v_len);
   char *buf = caml_stat_alloc(len);
-  memcpy(buf, String_val(v_buf) + Int_val(v_pos), len);
+  memcpy(buf, String_val(v_buf) + Long_val(v_pos), len);
   caml_enter_blocking_section();
-    v_res = Val_int(lo_write(conn, Int_val(v_fd), buf, len));
+    v_res = Val_long(lo_write(conn, Int_val(v_fd), buf, len));
     free(buf);
   caml_leave_blocking_section();
   CAMLreturn(v_res);
 }
 
+CAMLprim value lo_write_ba_stub(value v_conn, value v_fd,
+                                value v_buf, value v_pos, value v_len)
+{
+  CAMLparam2(v_conn, v_buf);
+  PGconn *conn = get_conn(v_conn);
+  value v_res;
+  size_t len = Long_val(v_len);
+  char *buf = ((char *) Caml_ba_data_val(v_buf)) + Long_val(v_pos);
+  caml_enter_blocking_section();
+    v_res = Val_long(lo_write(conn, Int_val(v_fd), buf, len));
+  caml_leave_blocking_section();
+  CAMLreturn(v_res);
+}
+
 CAMLprim value lo_lseek_stub(
   value v_conn, value v_fd, value v_pos, value v_whence)
 {
@@ -881,7 +1027,7 @@ CAMLprim value lo_lseek_stub(
       case 1 : whence = SEEK_CUR; break;
       default : whence = SEEK_END; break;
     }
-    v_res = Val_int(lo_lseek(conn, Int_val(v_fd), Int_val(v_pos), whence));
+    v_res = Val_int(lo_lseek(conn, Int_val(v_fd), Long_val(v_pos), whence));
   caml_leave_blocking_section();
   CAMLreturn(v_res);
 }
@@ -924,7 +1070,7 @@ CAMLprim value lo_import_stub(value v_conn, value v_fname)
   CAMLparam1(v_conn);
   PGconn *conn = get_conn(v_conn);
   value v_res;
-  int len = caml_string_length(v_fname) + 1;
+  size_t len = caml_string_length(v_fname) + 1;
   char *fname = caml_stat_alloc(len);
   memcpy(fname, String_val(v_fname), len);
   caml_enter_blocking_section();
@@ -939,7 +1085,7 @@ CAMLprim value lo_export_stub(value v_conn, value v_oid, value v_fname)
   CAMLparam1(v_conn);
   PGconn *conn = get_conn(v_conn);
   value v_res;
-  int len = caml_string_length(v_fname) + 1;
+  size_t len = caml_string_length(v_fname) + 1;
   char *fname = caml_stat_alloc(len);
   memcpy(fname, String_val(v_fname), len);
   caml_enter_blocking_section();
